/* Handles commands received by the server.  This does not necessarily
 * handle all commands - some might be in other files (like init.c)
 *
  * this file contains most of the commands for the dispatch loop most of
  * the functions are self-explanatory, the pixmap/bitmap commands recieve
  * the picture, and display it.  The drawinfo command draws a string
  * in the info window, the stats command updates the local copy of the stats
  * and displays it. handle_query prompts the user for input.
  * send_reply sends off the reply for the input.
  * player command gets the player information.
  * MapScroll scrolls the map on the client by some amount
  * MapCmd displays the map either with layer packing or stack packing. 
  *   packing/unpacking is best understood by looking at the server code
  *   (server/ericserver.c)
  *   stack packing is easy, for every map entry that changed, we pack
  *   1 byte for the x/y location, 1 byte for the count, and 2 bytes per
  *   face in the stack.
  *   layer packing is harder, but I seem to remember more efficient:
  *   first we pack in a list of all map cells that changed and are now
  *   empty.  The end of this list is a 255, which is bigger that 121, the
  *   maximum packed map location.  
  *   For each changed location we also pack in a list of all the faces and
  *   X/Y coordinates by layer, where the layer is the depth in the map.
  *   This essentially takes slices through the map rather than stacks.
  *   Then for each layer, (max is MAXMAPCELLFACES, a bad name) we start
  *   packing the layer into the message.  First we pack in a face, then
  *   for each place on the layer with the same face, we pack in the x/y
  *   location.  We mark the last x/y location with the high bit on 
  *   (11*11 = 121 < 128).  We then continue on with the next face, which
  *   is why the code marks the faces as -1 if they are finished.  Finally
  *   we mark the last face in the layer again with the high bit, clearly
  *   limiting the total number of faces to 32767, the code comments it's
  *   16384, I'm not clear why, but the second bit may be used somewhere
  *   else as well.
  *   The unpacking routines basically perform the opposite operations.
  */

#include <client.h>
#include <X11/Xlib.h>


/* Handles when the server says we can't be added.  In reality, we need to
 * close the connection and quit out, because the client is going to close
 * us down anyways.
 */
void AddMeFail(char *data, int len)
{
    fprintf(stderr,"addme_failed received.\n");
    return;
}

/* This is really a throwaway command - there really isn't any reason to
 * send addme_success commands.
 */
void AddMeSuccess(char *data, int len)
{
    fprintf(stderr,"addme_success received.\n");
    return;
}


void PixMapCmd(unsigned char *data,  int len)
{  
    int pnum,plen;

    pnum = GetInt_String(data);
    plen = GetInt_String(data+4);
/*    fprintf(stderr,"Recieved pixmap %d, len %d", pnum, plen);*/
    if (len<8 || (len-8)!=plen) {
	fprintf(stderr,"PixMapCmd: Lengths don't compare (%d,%d)\n",
		(len-8),plen);
	return;
    }
    display_newpixmap(pnum,data+8,plen);
}

void BitMapCmd(unsigned char *data, int len)
{
    int pnum = GetInt_String(data);

    /* We take off 6 bytes for image number, fg, bg */
    if ((len-6) != (24*3)) {
	fprintf(stderr,"Incorrect length on bitmap buffer should be %d was %d\n",
		24*3,len-6);
	return;
    }
    display_newbitmap(pnum,data[4],data[5],data+6);
}


void DrawInfoCmd(char *data, int len)
{
    int color=atoi(data);
    char *buf;

    buf = strchr(data, ' ');
    if (!buf) {
	fprintf(stderr,"DrawInfoCmd - got no data\n");
	buf="";
    }
    else buf++;
    if (color!=NDI_BLACK)
	draw_color_info(color, buf);
    else
	draw_info(buf,NDI_BLACK);

}

void StatsCmd(unsigned char *data, int len)
{
  int i=0;
  int c;

    while (i<len) {
	c=data[i++];
	switch (c) {
	    case CS_STAT_HP:	cpl.stats.hp=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_MAXHP:cpl.stats.maxhp=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_SP:	cpl.stats.sp=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_MAXSP:cpl.stats.maxsp=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_GRACE:cpl.stats.grace=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_MAXGRACE:cpl.stats.maxgrace=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_STR:	cpl.stats.Str=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_INT:	cpl.stats.Int=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_POW:	cpl.stats.Pow=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_WIS:	cpl.stats.Wis=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_DEX:	cpl.stats.Dex=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_CON:	cpl.stats.Con=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_CHA:  cpl.stats.Cha=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_EXP:  cpl.stats.exp=GetInt_String(data+i); i+=4; break;
	    case CS_STAT_LEVEL:cpl.stats.level=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_WC:   cpl.stats.wc=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_AC:   cpl.stats.ac=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_DAM:  cpl.stats.dam=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_ARMOUR:cpl.stats.armor=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_SPEED: cpl.stats.speed=GetInt_String(data+i); i+=4; break;
	    case CS_STAT_FOOD:  cpl.stats.food=GetShort_String(data+i); i+=2; break;
	    case CS_STAT_WEAP_SP:cpl.stats.weapon_sp=GetInt_String(data+i); i+=4; break;

	    case CS_STAT_RANGE: {
		int rlen=data[i++];
		strncpy(cpl.range,data+i,rlen);
		cpl.range[rlen]='\0';
		i += rlen;
		break;
	    }
	    case CS_STAT_TITLE: {
		int rlen=data[i++];
		strncpy(cpl.title,data+i,rlen);
		cpl.title[rlen]='\0';
		i += rlen;
		break;
	    }
	    default:
		fprintf(stderr,"Unknown stat number %d\n",c);
		abort();
	}
    }
    if (i>len) {
	fprintf(stderr,"got stats overflow, processed %d bytes out of %d\n",
		i, len);
    }
    draw_stats(0);
    draw_message_window(0);
}

void handle_query (char *data, int len)
{
    char *buf;
    uint8 flags = atoi(data);

    /* The actual text is optional */
    buf = strchr(data,' ');
    if (buf) buf++;

    /* If we just get passed an empty string, why draw this? */
    if (buf) draw_prompt(buf);

    /* Yes/no - don't do anything with it now */
    if (flags & CS_QUERY_YESNO) {}

    /* one character response expected */
    if (flags & CS_QUERY_SINGLECHAR)
	cpl.input_state = Reply_One;
    else
	cpl.input_state = Reply_Many;

    if (flags & CS_QUERY_HIDEINPUT)		/* no echo */
	cpl.no_echo=1;

      fprintf(stderr,"Received query.  Input state now %d\n", cpl.input_state);
}

/* Sends a reply to the server.  text contains the null terminated
 * string of text to send.  This function basically just packs
 * the stuff up.
 */

void send_reply(char *text)
{
    char buf[MAXSOCKBUF];

    sprintf(buf,"reply %s", text);

    cs_write_string(csocket.fd, buf, strlen(buf));
}



/* This function copies relevant data from the archetype to the
 * object.  Only copies data that was not set in the object
 * structure.
 *
 */

void PlayerCmd(unsigned char *data, int len)
{
    char name[MAX_BUF];
    int tag, weight, face,i=0,nlen;

    tag = GetInt_String(data);
    i+=4;
    weight = GetInt_String(data+i);
    i+=4;
    face= GetInt_String(data+i);
    i+=4;
    nlen=data[i++];
    strncpy(name, data+i, nlen);
    name[nlen]='\0';
    i+= nlen;

    if (i!=len) {
	fprintf(stderr,"PlayerCmd: lengths do not match (%d!=%d)\n",
		len, i);
    }
    new_player(tag, name, weight, face);
}

void item_actions (item *op)
{
    if (op->open) {
	open_container (op);
    } else if (op->was_open) {
	close_container (op);
    }
/*
    if (op->env == cpl.below) {
	check_auto_pickup (op);
    }
*/
}

/* ItemCmd grabs and display information for items in the inventory */
void ItemCmd(unsigned char *data, int len)
{
    int weight, loc, tag, face, flags,pos=0,nlen;
    char name[MAX_BUF];

    loc = GetInt_String(data);
    pos+=4;

    if (pos == len) {
	/* An empty item command can be used to delete the whole
	 * inventory of item 'loc'. Esspecially if loc is 0, then
	 * there is no floor under the player.
	 */
/*	delete_item_inventory(loc);*/
	fprintf(stderr,"ItemCmd: Got location with no other data\n");
    }
    else if (loc < 0) { /* delele following items */
	fprintf(stderr,"ItemCmd: got location with negative value (%d)\n", loc);
	return;
    } else {
	while (pos < len) {
	    tag=GetInt_String(data+pos); pos+=4;
	    flags = GetInt_String(data+pos); pos+=4;
	    weight = GetInt_String(data+pos); pos+=4;
	    face = GetInt_String(data+pos); pos+=4;
	    nlen = data[pos++];
	    strncpy(name, (char*)data+pos, nlen);
	    pos += nlen;
	    name[nlen]='\0';
	    update_item (tag, loc, name, weight, face, flags);
	    item_actions (locate_item(tag));
	}
	if (pos>len) 
	    fprintf(stderr,"ItemCmd: Overread buffer: %d > %d\n", pos, len);
    }
}

void DeleteItem(unsigned char *data, int len)
{
    int pos=0,tag;

    while (pos<len) {
	tag=GetInt_String(data); pos+=4;
	delete_item(tag);
    }
    if (pos>len) 
	fprintf(stderr,"ItemCmd: Overread buffer: %d > %d\n", pos, len);
}

void DeleteInventory(unsigned char *data, int len)
{
    int tag;

    tag=atoi(data);
    if (tag<0) {
	fprintf(stderr,"DeleteInventory: Invalid tag: %d\n", tag);
	return;
    }
    remove_item_inventory(locate_item(tag));
}

void Map_unpacklayer(unsigned char *cur,unsigned char *end)
{
  long x,y,face;
  int xy;
  int clear = 1;
  unsigned char *fcur;
  
  while (cur < end && *cur != 255) {
    xy = *cur;
    cur++;
    x = xy / 11;
    y = xy % 11;
    display_map_clearcell(x,y);
  }
  cur++;
  while(cur < end) {
    fcur = cur;
    face = *cur & 0x7f;
    cur++;
    face = (face << 8) | *cur;
    cur++;

    /* Now process all the spaces with this face */
    while(1) {
      xy = *cur & 0x7f;
      x = xy / 11;
      y = xy % 11;
      if (clear) {
	display_map_clearcell(x,y);
      }
      display_map_addbelow(x,y,face);
      xy = *cur;
      cur++;
      if (xy & 128) {
	break;
      }
      if (cur==end) {
	fprintf(stderr,"Incorrectly packed map.\n");
	abort();
      }
    }
    /* Got end of layer */
    if (*fcur & 128) 
      clear = 0; /* end of first layer (or some layer after the first) */
  }
}

void MapCmd(unsigned char *data, int len)
{
    unsigned char *end;
  
    end = data + len;

    display_map_startupdate();
    Map_unpacklayer(data,end);
    display_map_doneupdate();
}

void map_scrollCmd(char *data, int len)
{
    int dx,dy;
    char *buf;

    dx = atoi(data);
    buf = strchr(data,' ');
    if (!buf) {
	fprintf(stderr,"map_scrollCmd: Got short packet.\n");
	return;
    }
    buf++;
    dy = atoi(buf);
    display_mapscroll(dx,dy);
}

void look_at(int x, int y)
{
    fprintf(stderr,"look_at %d %d not implemented.\n",x,y);
}
